// SPDX-License-Identifier: MIT OR Apache-2.0

package outputs

import (
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"strconv"

	"github.com/DataDog/datadog-go/statsd"
	"github.com/google/uuid"
	"go.opentelemetry.io/otel/attribute"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/tools/clientcmd"

	"github.com/falcosecurity/falcosidekick/internal/pkg/utils"
	otlpmetrics "github.com/falcosecurity/falcosidekick/outputs/otlp_metrics"
	"github.com/falcosecurity/falcosidekick/types"
)

// Some constant strings to use in request headers
const FissionEventIDKey = "event-id"
const FissionEventNamespaceKey = "event-namespace"
const FissionContentType = "application/json"
const APIv1Namespaces = "/api/v1/namespaces/"
const ServicesPath = "/services/"

// NewFissionClient returns a new output.Client for accessing Kubernetes.
func NewFissionClient(config *types.Configuration, stats *types.Statistics, promStats *types.PromStatistics,
	otlpMetrics *otlpmetrics.OTLPMetrics, statsdClient, dogstatsdClient *statsd.Client) (*Client, error) {
	if config.Fission.KubeConfig != "" {
		restConfig, err := clientcmd.BuildConfigFromFlags("", config.Fission.KubeConfig)
		if err != nil {
			return nil, err
		}
		clientset, err := kubernetes.NewForConfig(restConfig)
		if err != nil {
			return nil, err
		}
		return &Client{
			OutputType:       Fission,
			Config:           config,
			Stats:            stats,
			PromStats:        promStats,
			OTLPMetrics:      otlpMetrics,
			StatsdClient:     statsdClient,
			DogstatsdClient:  dogstatsdClient,
			KubernetesClient: clientset,
			cfg:              config.Fission.CommonConfig,
		}, nil
	}

	endpointUrl := fmt.Sprintf("http://%s.%s.svc.cluster.local:%d/fission-function/%s", config.Fission.RouterService, config.Fission.RouterNamespace, config.Fission.RouterPort, config.Fission.Function)
	initClientArgs := &types.InitClientArgs{
		Config:          config,
		Stats:           stats,
		DogstatsdClient: dogstatsdClient,
		PromStats:       promStats,
		StatsdClient:    statsdClient,
	}

	return NewClient(Fission, endpointUrl, config.Fission.CommonConfig, *initClientArgs)
}

// FissionCall .
func (c *Client) FissionCall(falcopayload types.FalcoPayload) {
	c.Stats.Fission.Add(Total, 1)

	if c.Config.Fission.KubeConfig != "" {
		str, _ := json.Marshal(falcopayload)
		req := c.KubernetesClient.CoreV1().RESTClient().Post().AbsPath(APIv1Namespaces +
			c.Config.Fission.RouterNamespace + ServicesPath + c.Config.Fission.RouterService +
			":" + strconv.Itoa(c.Config.Fission.RouterPort) + "/proxy/" + "/fission-function/" +
			c.Config.Fission.Function).Body(str)
		req.SetHeader(FissionEventIDKey, uuid.New().String())
		req.SetHeader(ContentTypeHeaderKey, FissionContentType)
		req.SetHeader(UserAgentHeaderKey, UserAgentHeaderValue)

		res := req.Do(context.TODO())
		rawbody, err := res.Raw()
		if err != nil {
			go c.CountMetric(Outputs, 1, []string{"output:Fission", "status:error"})
			c.Stats.Fission.Add(Error, 1)
			c.PromStats.Outputs.With(map[string]string{"destination": "Fission", "status": Error}).Inc()
			c.OTLPMetrics.Outputs.With(attribute.String("destination", "Fission"),
				attribute.String("status", Error)).Inc()
			utils.Log(utils.ErrorLvl, c.OutputType, err.Error())
			return
		}
		utils.Log(utils.InfoLvl, c.OutputType, fmt.Sprintf("Function Response : %v", string(rawbody)))
	} else {
		c.ContentType = FissionContentType

		err := c.Post(falcopayload, func(req *http.Request) {
			req.Header.Set(FissionEventIDKey, uuid.New().String())
		})
		if err != nil {
			go c.CountMetric(Outputs, 1, []string{"output:Fission", "status:error"})
			c.Stats.Fission.Add(Error, 1)
			c.PromStats.Outputs.With(map[string]string{"destination": "Fission", "status": Error}).Inc()
			c.OTLPMetrics.Outputs.With(attribute.String("destination", "Fission"),
				attribute.String("status", Error)).Inc()
			utils.Log(utils.ErrorLvl, c.OutputType, err.Error())
			return
		}
	}
	utils.Log(utils.InfoLvl, c.OutputType, fmt.Sprintf("Call Function \"%v\" OK", c.Config.Fission.Function))
	go c.CountMetric(Outputs, 1, []string{"output:Fission", "status:ok"})
	c.Stats.Fission.Add(OK, 1)
	c.PromStats.Outputs.With(map[string]string{"destination": "Fission", "status": OK}).Inc()
	c.OTLPMetrics.Outputs.With(attribute.String("destination", "Fission"),
		attribute.String("status", OK)).Inc()
}
